"""
"""

__all__ = ["Value", "Attribute"]

import dill
import numpy as np

from collections import Iterable
from copy import copy
from typing import Any, Dict, List, Tuple, Union

from .annotation import Annotation, AnnotationResult
from .invariants import invariant


class Value(Annotation):
    """
    Annotation class for asserting that a value should be observed.

    Indicates that a value passed to the constructor should be observed while tracing through the
    students' code. Values can be of any type that is picklable by ``dill``. Values can specify a
    list of :ref:`invariants<invariants>` that will allow objects to be considered "equal." For 
    values that support arithemtic operators, absolute tolerances can be specified as well.

    Args:
        value (``object``): the value that should be observed
        tol (``float`` or ``int``, optional): absolute tolerance for numeric values
        invariants (``list[invariant]``): invariants for 
            this value
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    intial_value: Any
    """a copy of the value passed to the constructor"""

    _values: List[Any]
    """
    a list of values that resulted from passing the intial value through the series of invariants 
    specified in the constructor
    """

    tol: Union[float, int]
    """absolute tolerance for numeric values"""

    invariants: List[invariant]
    """the invariants for this value"""

    def __init__(self, value: Any, tol: Union[float, int] = 0, invariants: List[invariant] = [], **kwargs):
        try:
            dill.dumps(value)
        except Exception as e:
            raise ValueError(f"Values must be serializable but the following error was thrown during serialization:\n{e}")

        self.initial_value = copy(value)
        self._values = [self.initial_value]
        self.tol = tol
        self.invariants = invariants

        for inv in self.invariants:
            self._values = inv(self._values)

        super().__init__(**kwargs)

    @property
    def children(self):
        return []

    def check(self, observed_values: List[Tuple[Any, float]]) -> AnnotationResult:
        """
        Checks that the value tracked by this annotation occurs in the list of observed values.

        Checks that the value under the conditions of its invariants occurs in the list of tuples of
        observed values and timestamps ``observed_values``. Creates and returns an 
        :py:class:`AnnotationResult<pybryt.AnnotationResult>` object with the results of this check.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        satisfied = [self._check_observed_value(v) for v in observed_values]
        if not any(satisfied):
            return AnnotationResult(False, self)

        first_satisfier = satisfied.index(True)
        return AnnotationResult(True, self, observed_values[first_satisfier][0], observed_values[first_satisfier][1])

    def _check_observed_value(self, observed_value: Tuple[Any, float]) -> bool:
        """
        Checks whether a single observed value tuple satisfies this value.

        Applies all invariants to the first element of the tuple and then checks whether any of the
        resulting values match and of the values in ``self._values``.

        Args:
            observed_value (``tuple[object, float]``): the observed value tuple

        Returns:
            ``bool``: whether the value matched
        """
        other_values = [observed_value[0]]
        for inv in self.invariants:
            other_values = inv(other_values)
        
        for value in self._values:
            for other_value in other_values:
                # if type(value) != type(other_value):
                #     continue

                # else:
                if hasattr(value, 'shape'):
                    try:
                        if value.shape != other_value.shape:
                            continue
                    except AttributeError:
                        continue

                try:
                    ub, lb = value + self.tol, value - self.tol
                    numeric = True
                
                except:
                    numeric = False

                if numeric:
                    try:
                        res = np.logical_and(ub >= other_value, other_value >= lb)
                    except (ValueError, TypeError) as e:
                        continue
                else:
                    try:
                        res = value == other_value
                    except (ValueError, TypeError) as e:
                        continue

                if isinstance(res, Iterable):
                    try:
                        res = all(res)
                    except ValueError as e:
                        if isinstance(res, np.ndarray):
                            res = res.all()
                        else:
                            raise e

                if res:
                    return True

        return False


class _AttrValue(Value):
    """
    Wrapper around ``Value`` for checking whether an object has an *attribute* matching a specific
    value.

    Args:
        obj (``object``): the object being checked
        attr (``str``): the attribute of the object being checked
        **kwargs: additional keyword arguments passed to the :py:class:`Value<pybryt.Value>` 
            constructor
    """

    _object: Any
    _attr: str

    def __init__(self, obj: Any, attr: str, **kwargs):
        self._object = obj
        self._attr = attr
        val = getattr(obj, attr)
        super().__init__(val, **kwargs)
    
    def check(self, observed_values: List[Tuple[Any, float]]) -> AnnotationResult:
        """
        Checks whether any of the values in ``observed_values`` has an attribute matching the value
        in this annotation.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        vals = [t for t in observed_values if hasattr(t[0], self._attr)]
        attrs = [(getattr(obj, self._attr), t) for obj, t in vals]
        res = super().check(attrs)
        satisfier = vals[attrs.index(res.value)]
        return AnnotationResult(None, self, value=satisfier, children=[res])


class Attribute(Annotation):
    """
    Annotation for asserting that an object has an attribute value matching that of another object.

    Uses a :py:class:`Value<pybryt.Value>` annotation to check there exists an object in the 
    students' code which as a specific attribute, or a series of attributes, with values matching
    those in the object passed to the constructor. The constructor taks in the object in question
    and the name of the attribute(s) being studied.

    Args:
        obj (``object``): the object being checked
        attrs (``str`` or ``list[str]``): the attribute or attributes that should be checked
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    _annotations: List[_AttrValue]

    def __init__(self, obj: Any, attrs: Union[str, List[str]], **kwargs):
        if isinstance(attrs, str):
            attrs = [attrs]
        if not isinstance(attrs, list) and not all(isinstance(a, str) for a in attrs):
            raise TypeError(f"Invalid type for argument 'attrs': {type(attrs)}")
                
        self._annotations = []
        for attr in attrs:
            if not hasattr(obj, attr):
                raise AttributeError(f"{obj} has not attribute '{attr}'")
            self._annotations.append(_AttrValue(obj, attr, **kwargs))

    @property
    def children(self):
        return self._annotations

    def check(self, observed_values: List[Tuple[Any, float]]) -> AnnotationResult:
        """
        Checks whether any of the values in ``observed_values`` has all of the required attributes,
        each matching the values expected.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = [v.check(observed_values) for v in self._annotations]        
        return AnnotationResult(None, self, children=results)



"""
"""

import os
import re
import time
import tempfile
import linecache
import dill
import nbformat

from nbconvert.preprocessors import ExecutePreprocessor
from copy import copy
from types import FrameType, FunctionType, ModuleType
from typing import Any, List, Tuple, Callable, Optional
from textwrap import dedent

from .preprocessors import IntermediateVariablePreprocessor
from .utils import make_secret, pickle_and_hash


NBFORMAT_VERSION = 4


# class ObservedValue:
#     """
#     """

#     timestamp: float
#     value: Any
#     fn_name: str
#     filename: str

#     def __init__(self, value: Any, fn_name: str = None, filename: str = None, timestamp: float = None):
#         self.value = value
#         self.fn_name = fn_name
#         self.filename = filename
#         self.timestamp = timestamp


def create_collector(skip_types: List[type] = [type, type(len), ModuleType, FunctionType], addl_filenames: List[str] = []) -> \
        Tuple[List[Tuple[Any, float]], Callable[[FrameType, str, Any], Callable]]:
    """
    """
    observed = []
    vars_not_found = {}
    hashes = set()
    # prune_pickup_idx = [0]
    # next_prune = [PRUNE_FREQUENCY]

    # def prune_observations():
    #     to_delete = []
    #     for i in range(prune_pickup_idx[0], len(observed)):
    #         try:
    #             h = pickle_and_hash(observed[i])
    #             if h in hashes[0]:
    #                 to_delete.append(i)
    #             else:
    #                 hashes[0] |= set([h])
    #         except:
    #             to_delete.append(i)
        
    #     to_delete.reverse()
    #     for i in to_delete:
    #         observed.pop(i)
        
    #     prune_pickup_idx[0] = len(observed)

    def track_value(val, seen_at):
        if type(val) in skip_types:
            return

        try:
            h = pickle_and_hash(val)
        except:
            return
        
        if h not in hashes:
            observed.append((copy(val), seen_at))
            hashes.add(h)

    # TODO: a way to track the cell of execution
    def collect_intermidiate_results(frame: FrameType, event: str, arg: Any):
        seen_at = time.time()

        name = frame.f_code.co_filename + frame.f_code.co_name
        
        if frame.f_code.co_filename.startswith("<ipython") or frame.f_code.co_filename in addl_filenames:
            if event == "line" or event == "return":

                line = linecache.getline(frame.f_code.co_filename, frame.f_lineno)
                tokens = set("".join(char if char.isalnum() or char == '_' else "\n" for char in line).split("\n"))
                for t in "".join(char if char.isalnum() or char == '_' or char == '.' else "\n" for char in line).split("\n"):
                    tokens.add(t)

                # for tracking the results of an assignment statement
                m = re.match(r"\s*(\w+)\s=.*", line)
                if m:
                    if name not in vars_not_found:
                        vars_not_found[name] = []
                    vars_not_found[name].append((m.group(1), seen_at))
                
                for t in tokens:
                    if "." in t:
                        try:
                            val = eval(t, frame.f_globals, frame.f_locals)
                            track_value(val, seen_at)
                        except:
                            pass

                    else:
                        if t in frame.f_locals:
                            val = frame.f_locals[t]
                            track_value(val, seen_at)
                                
                        elif t in frame.f_globals:
                            val = frame.f_globals[t]
                            track_value(val, seen_at)

            if event == "return" and type(arg) not in skip_types:
                track_value(arg, seen_at)

        elif (frame.f_back.f_code.co_filename.startswith("<ipython") or \
                frame.f_back.f_code.co_filename in addl_filenames) and event == "return":
            track_value(arg, seen_at)

        if event == "return" and name in vars_not_found:
            varnames = vars_not_found.pop(name)
            for t, timestamp in varnames:
                if t in frame.f_locals:
                    val = frame.f_locals[t]
                    track_value(val, timestamp)

                elif t in frame.f_globals:
                    val = frame.f_globals[t]
                    track_value(val, timestamp)
        
        # if len(observed) >= next_prune[0]:
        #     prune_observations()
        #     next_prune[0] = len(observed) + PRUNE_FREQUENCY

        return collect_intermidiate_results

    return observed, collect_intermidiate_results


def execute_notebook(nb: nbformat.NotebookNode, addl_filenames: List[str] = [], output: Optional[str] = None) -> \
        Tuple[float, float, List[Tuple[Any, float]]]:
    """
    """
    preprocessor = IntermediateVariablePreprocessor()
    nb = preprocessor.preprocess(nb)

    secret = make_secret()
    _, observed_fp = tempfile.mkstemp()

    first_cell = nbformat.v4.new_code_cell(dedent(f"""\
        import sys
        from pybryt.execution import create_collector
        observed_{secret}, cir = create_collector(addl_filenames={addl_filenames})
        sys.settrace(cir)
    """))

    last_cell = nbformat.v4.new_code_cell(dedent(f"""\
        sys.settrace(None)
        import dill
        from pybryt.utils import filter_pickleable_list
        filter_pickleable_list(observed_{secret})
        with open("{observed_fp}", "wb+") as f:
            dill.dump(observed_{secret}, f)
    """))

    nb['cells'].insert(0, first_cell)
    nb['cells'].append(last_cell)

    ep = ExecutePreprocessor(timeout=1200, allow_errors=True)

    execution_start = time.time()
    ep.preprocess(nb)
    execution_end = time.time()

    if output:
        with open(output, "w+") as f:
            nbformat.write(nb, f)

    with open(observed_fp, "rb") as f:
        observed = dill.load(f)

    os.remove(observed_fp)

    return execution_start, execution_end, observed



"""
Tests for PyBryt annotations
"""

import time
import numpy as np

from collections.abc import Iterable
from functools import lru_cache

from pybryt import Value
from pybryt.utils import pickle_and_hash


START_TIMESTAMP = 1614904732.51892


@lru_cache(1)
def generate_memory_footprint():
    """
    """
    np.random.seed(42)    
    return [
        (np.random.uniform(-100, 100, size=(100, 100)), time.time()),
        (4.0, time.time()),
        (list(range(100))[::-1], time.time()),
        (1, time.time()),
        (np.e, time.time()),
        (None, time.time()),
        (None, time.time()),
        (np.random.normal(size=102), time.time()),
        (4.0, time.time()),
    ]


def test_value_annotation():
    """
    """
    mfp = generate_memory_footprint()

    seen = {}
    for val, ts in mfp:
        v = Value(val)
        res = v.check(mfp)

        h = pickle_and_hash(val)

        # check attributes of values and results
        assert len(v.children) == 0, "Value annotation has children"
        assert res.satisfied is True, "Did not find value in memory footprint"
        assert res._satisfied is True, "Did not find value in memory footprint"
        assert res.annotation is v, "Wrong annotation in result"
        assert res.children is None, "Value annotation result has children"
        if h in seen:
            # check that we get the earliest timestamp for duplicate values
            assert np.isclose(res.timestamp, seen[h]), \
                "Wrong timestamp for duplicate value in value annotation result"
        else:
            assert np.isclose(res.timestamp, ts), "Wrong timestamp in value annotation result"
        
        if isinstance(val, Iterable) and hasattr(val, "all"): # for numpy arrays
            assert (res.value == val).all(), "Wrong value in value annotation result"
        else:
            assert res.value == val, "Wrong value in value annotation result"

        if h not in seen:
            seen[h] = ts

    v = Value(-1) # does not occur in mfp
    res = v.check(mfp)

    # check attributes of values and results
    assert len(v.children) == 0, "Value annotation has children"
    assert res.satisfied is False, "Did not find value in memory footprint"
    assert res._satisfied is False, "Did not find value in memory footprint"
    assert res.annotation is v, "Wrong annotation in result"
    assert res.children is None, "Value annotation result has children"
    assert res.timestamp == -1, "Wrong timestamp in value annotation result"
    assert res.value is None, "Wrong value in value annotation result"



"""
"""

import os
import dill
import base64
import nbformat

from typing import Any, List, NoReturn, Optional, Tuple, Union

from .reference import ReferenceImplementation, ReferenceResult
from .execution import execute_notebook


NBFORMAT_VERSION = 4


class StudentImplementation:
    """
    """

    nb: nbformat.NotebookNode
    values: List[Tuple[Any, float]]
    start: float
    end: float

    def __init__(
        self, path_or_nb: Union[str, nbformat.NotebookNode], addl_filenames: List[str] = [],
        output: Optional[str] = None
    ):
        if isinstance(path_or_nb, str):
            self.nb = nbformat.read(path_or_nb, as_version=NBFORMAT_VERSION)
        elif isinstance(path_or_nb, nbformat.NotebookNode):
            self.nb = path_or_nb
        else:
            raise ValueError(f"path_or_nb is of unsupported type {type(path_or_nb)}")

        self._execute(addl_filenames=addl_filenames, output=output)

    def _execute(self, addl_filenames: List[str] = [], output: Optional[str] = None) -> NoReturn:
        self.start, self.end, self.values = execute_notebook(self.nb, addl_filenames=addl_filenames, output=output)

    def dump(self, dest: str = "student.pkl") -> NoReturn:
        with open(dest, "wb+") as f:
            dill.dump(self, f)

    def dumps(self) -> str:
        bits = dill.dumps(self)
        return base64.b64encode(bits).decode("ascii")

    @staticmethod
    def load(file: str) -> Union['StudentImplementation']:
        with open(file, "rb") as f:
            instance = dill.load(f)
        return instance

    @classmethod
    def loads(cls, data: str) -> "StudentImplementation":
        return dill.loads(base64.b64decode(data.encode("ascii")))

    def check(self, ref: Union[ReferenceImplementation, List[ReferenceImplementation]], group=None) -> \
            Union[ReferenceResult, List[ReferenceResult]]:
        if isinstance(ref, ReferenceImplementation):
            return ref.run(self.values, group=group)
        elif isinstance(ref, list):
            return [r.run(self.values, group=group) for r in ref]
        else:
            raise ValueError(f"check cannot take values of type '{type(ref)}'")
    
    def check_plagiarism(self, student_impls: List["StudentImplementation"], **kwargs) -> List[ReferenceResult]:
        refs = create_references([self], **kwargs)
        return get_impl_results(refs[0], student_impls, **kwargs)


from .plagiarism import create_references, get_impl_results



"""Invariants for value annotations (available as :py:mod:`pybryt.invariants`)"""

import numpy as np

from abc import ABC, abstractmethod
from collections import Iterable
from typing import Any, List, Optional, Union
# from enum import Enum, auto

# TODO: add iterable_type invariant


class invariant(ABC):
    """
    Abstract base class for invariants. 
    
    All subclasses should implement the :py:meth:`run<invariant.run>` static method for generating 
    values that this invariant accepts as "correct". Invariants have a custom ``__new__`` method 
    that returns the value of calling the :py:meth:`run<invariant.run>` method, making them 
    function as callables.
    """

    @staticmethod
    def __new__(cls, *args, **kwargs):
        return cls.run(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def run(values: List[Any], **kwargs) -> List[Any]:
        """
        Returns a list of values that this invariant accepts as correct.

        Takes in a list of acceptable values from a :py:class:`Value<pybryt.Value>` annotation and
        returns a list of values that would evaluate as "the same" under the conditions of this 
        invariant.

        For example, if ``values`` as a list with a single element, a numpy matrix, and the 
        invariant was matrix transposition, this method would return a length-2 list containing the
        original matrix and its transpose.

        Args:
            values (``list[object]``): acceptable values, either from the initial constructor call
                of the annotation or from the results of other invariants
            kwargs: additional keyword arguments
        
        Returns:
            ``list[object]``: the values that would evaluate as "the same" under the conditions 
            of this invariant
        """
        ...


# TODO: if hashing, for all strings collect actual string and lowercased version (marked as such), 
#       and compare against that if this invariant is used. 
class string_capitalization(invariant):
    """
    An invariant that compares strings ignoring capitalization.

    Works by taking in a list of values and lowercasing them if they are strings and leaving them
    unchanged otherwise.
    """

    @staticmethod
    def run(values: List[Any]) -> List[Any]:
        """
        Returns a list of values in which all strings have been lowercased.

        Args:
            values (``list[object]``): acceptable values, either from the initial constructor call
                of the annotation or from the results of other invariants
        
        Returns:
            ``list[object]``: the elements of ``values`` with all strings lowercased
        """
        ret = []
        for v in values:
            if not isinstance(v, str):
                ret.append(v)
            else:
                ret.append(v.lower())
        return ret



import ast
import astunparse
import nbformat

from IPython.core.inputtransformer2 import TransformerManager

from ..utils import make_secret, notebook_to_string


def get_varname():
    return f"var_{make_secret()}"


def add_parents(root):
    for node in ast.walk(root):
        for child in ast.iter_child_nodes(node):
            child.parent = node


class UnassignedVarWrapper(ast.NodeTransformer):

    _skip_node_types = [ast.ListComp, ast.SetComp, ast.GeneratorExp, ast.DictComp, ast.Lambda]
    
    def __init__(self):
        super().__init__()
        self.insertions = [] # list of (parent, attr, idx, node)
    
    def visit(self, root, *args, **kwargs):
        self.root = root
        ret = super().visit(root, *args, **kwargs)
        for parent, attr, idx, node in self.insertions:
            getattr(parent, attr)[idx] = node
        return ret

    def transform_unassigned_node(self, node):
        if not isinstance(node.parent, ast.Assign):
            vn = get_varname()
            curr = node.parent
            body_child = None
            while not isinstance(curr, ast.Module):#hasattr(curr.parent, "body"):
                if hasattr(curr.parent, "body") and not (hasattr(curr.parent, "test") and \
                        curr.parent.test == curr) and body_child is None:
                    body_child = curr
                
                # don't perform if in a comprehension
                if any(isinstance(curr, t) for t in type(self)._skip_node_types):
                    # for n in ast.iter_child_nodes(node):
                    #     self.visit(n)
                    return node
                
                curr = curr.parent
            
            curr = body_child
            is_else = False
            try:
                idx = curr.parent.body.index(curr)
            except ValueError:
                if isinstance(curr.parent, ast.If):
                    idx = curr.parent.orelse.index(curr)
                    is_else = True
                else:
                    raise
            
            curr = curr.parent
            new_assign = ast.Assign([ast.Name(vn, ctx=ast.Load())], node, ctx=ast.Load())
            new_assign.parent = curr
            if is_else:
                curr.orelse.insert(idx, new_assign)
                self.insertions.append((curr, "orelse", idx, new_assign))
            else:
                curr.body.insert(idx, new_assign)
                self.insertions.append((curr, "body", idx, new_assign))

            new_name = ast.Name(vn, ctx=ast.Load())
            new_name.parent = node.parent
            node.parent = new_assign

            self.visit(new_assign)

            return new_name
        else:
            for n in ast.iter_child_nodes(node):
                self.visit(n)
        return node
    
    def visit_Call(self, node):
        return self.transform_unassigned_node(node)
    
    def visit_BinOp(self, node):
        return self.transform_unassigned_node(node)


class IntermediateVariablePreprocessor():

    def preprocess(self, nb):
        # code = notebook_to_string(nb)
        transformer_mgr = TransformerManager()
        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                code = cell['source']
                code = transformer_mgr.transform_cell(code)
                tree = ast.parse(code)
                add_parents(tree)
                transformer = UnassignedVarWrapper()
                tree = transformer.visit(tree)
                tree = ast.fix_missing_locations(tree)
                code = astunparse.unparse(tree)
                cell['source'] = code

        # nb = nbformat.v4.new_notebook()
        # nb['cells'].append(nbformat.v4.new_code_cell(code))
        return nb



"""Relational annotations for asserting conditions on other annotations"""

__all__ = [
    "RelationalAnnotation", "BeforeAnnotation", "AndAnnotation", "OrAnnotation", "XorAnnotation", 
    "NotAnnotation"
]

from abc import abstractmethod
from typing import Any, List, Tuple

from .annotation import Annotation, AnnotationResult


class RelationalAnnotation(Annotation):
    """
    Abstract base class for annotations that assert some kind of condition (temporal or boolean)
    on one or more other annotations.

    All relational annotations should inherit from this class, which defines a constructor that
    takes in a list of annotations and any necessary keyword arguments and populates the fields 
    needed for tracking child annotations.

    Args:
        *annotations (:py:class:`Annotation<pybryt.Annotation>`): the child annotations being 
            operated on
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    _annotations: List['Annotation']
    """the child annotations that this annotation operates on"""

    def __init__(self, *annotations, **kwargs):
        self._annotations = annotations
        for ann in self._annotations:
            if not isinstance(ann, Annotation):
                raise ValueError("One of the arguments is not an annotation")

        super().__init__(**kwargs)

    @property
    def children(self):
        """
        ``list[Annotation]``: the child annotations that this annotation operates on
        """
        return self._annotations

    @abstractmethod
    def check(self, observed_values: List[Tuple[Any, float]]) -> "AnnotationResult":
        ...


class BeforeAnnotation(RelationalAnnotation):
    """
    Annotation for asserting that one annotation occurs before another.

    When being :py:meth:`check<pybryt.BeforeAnnotation.check>` is called, ensures that all child 
    annotations are satisfied and then checks that for the :math:`i^\\text{th}` annotation the
    :math:`(i+1)^\\text{th}` annotation has a timestamp greater than or equal to its own. 
    Annotations must be passed to the constructor in the order in which they are expected to appear.

    Args:
        *annotations (:py:class:`Annotation<pybryt.Annotation>`): the child annotations being 
            operated on
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    def check(self, observed_values: List[Tuple[Any, float]]) -> "AnnotationResult":
        """
        Checks that all child annotations are satisfied by the values in ``observed_values`` and
        that the timestamps of the satisfying values occur in non-decreasing order.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = []
        for ann in self._annotations:
            results.append(ann.check(observed_values))

        if all(res.satisfied for res in results):
            before = []
            for i in range(len(self._annotations) - 1):
                before.append(results[i].satisfied_at < results[i + 1].satisfied_at)
            
            return AnnotationResult(all(before), self, children = results)
        
        else:
            return AnnotationResult(False, self, children = results)


class AndAnnotation(RelationalAnnotation):
    """
    Annotation for asserting that a series of annotations are **all** satisfied.

    Args:
        *annotations (:py:class:`Annotation<pybryt.Annotation>`): the child annotations being 
            operated on
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    def check(self, observed_values: List[Tuple[Any, float]]) -> "AnnotationResult":
        """
        Checks that all child annotations are satisfied by the values in ``observed_values``.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = []
        for ann in self._annotations:
            results.append(ann.check(observed_values))

        return AnnotationResult(all(res.satisfied for res in results), self, children = results)


class OrAnnotation(RelationalAnnotation):
    """
    Annotation for asserting that, of a series of annotations, **any** are satisfied.

    Args:
        *annotations (:py:class:`Annotation<pybryt.Annotation>`): the child annotations being 
            operated on
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    def check(self, observed_values: List[Tuple[Any, float]]) -> "AnnotationResult":
        """
        Checks that any of the child annotations are satisfied by the values in ``observed_values``.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = []
        for ann in self._annotations:
            results.append(ann.check(observed_values))

        return AnnotationResult(any(res.satisfied for res in results), self, children = results)


class XorAnnotation(RelationalAnnotation):
    """
    Annotation for asserting that, of two annotations, one is satisfied and the other is not.

    Args:
        *annotations (:py:class:`Annotation<pybryt.Annotation>`): the child annotations being 
            operated on
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    def __init__(self, *annotations):
        super().__init__(*annotations)
        assert len(self._annotations) == 2, "Cannot use xor with more than two annotations"

    def check(self, observed_values: List[Tuple[Any, float]]) -> "AnnotationResult":
        """
        Checks that one child annotation is satisfied and one is not by the values in 
        ``observed_values``.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = []
        for ann in self._annotations:
            results.append(ann.check(observed_values))

        sats = [res.satisfied for res in results]
        return AnnotationResult(sats[0] ^ sats[1], self, children = results)


class NotAnnotation(RelationalAnnotation):
    """
    Annotation for asserting that a single annotation should **not** be satisfied.

    Args:
        *annotations (:py:class:`Annotation<pybryt.Annotation>`): the child annotation being 
            operated on
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    def check(self, observed_values: List[Tuple[Any, float]]) -> "AnnotationResult":
        """
        Checks that the child annotation is not satisfied by the values in ``observed_values``.

        Args:
            observed_values (``list[tuple[object, float]]``): a list of tuples of values observed
                during execution and the timestamps of those values
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation based on 
            ``observed_values``
        """
        results = []
        for ann in self._annotations:
            results.append(ann.check(observed_values))

        return AnnotationResult(all(not res.satisfied for res in results), self, children = results)



"""PyBryt plugin for Otter-Grader"""

import os
import dill
import shutil
import base64
import pathlib
import itertools
import warnings
import nbformat

from glob import glob
from otter.plugins import AbstractOtterPlugin
from otter.test_files import GradingResults
from otter.utils import get_source
from typing import List, NoReturn, Union

from .. import ReferenceImplementation, StudentImplementation
from ..execution import NBFORMAT_VERSION
from ..utils import save_notebook


class OtterPlugin(AbstractOtterPlugin):
    """
    Plugin for Otter-Grader.
    """

    IMPORTABLE_NAME = "pybryt.integrations.otter.OtterPlugin"
    _cached_refs = None
    _generate_report = None
    _student_impl = None

    def during_assign(self, assignment):
        self._cached_refs = []
        for fn in self.plugin_config["references"]:
            ref = ReferenceImplementation.compile(fn)
            if not isinstance(ref, list):
                ref = [ref]
            
            self._cached_refs.extend(ref)
            
            agp = assignment.result / 'autograder' / (pathlib.Path(fn).stem + ".pkl")
            with open(agp, "wb+") as f:
                dill.dump(ref, f)
            
            stp = assignment.result / 'student' / (pathlib.Path(fn).stem + ".pkl")
            shutil.copy(str(agp), str(stp))

    def during_generate(self, otter_config, assignment):
        if assignment is not None:
            cwd = os.getcwd()
            os.chdir(assignment.master.parent)

        cfg_idx = [self.IMPORTABLE_NAME in c.keys() if isinstance(c, dict) else False for c in otter_config["plugins"]].index(True)
        if self._cached_refs is None:
            refs = []
            for fn in otter_config["plugins"][cfg_idx][self.IMPORTABLE_NAME]["references"]:
                ref = ReferenceImplementation.compile(fn)
                if not isinstance(ref, list):
                    ref = [ref]
                refs.extend(ref)
        else:
            # if during_assign was run, this should be the same
            refs = self._cached_refs
        
        refs = base64.b64encode(dill.dumps(refs)).decode("ascii")
        otter_config["plugins"][cfg_idx][self.IMPORTABLE_NAME]["reference_bytes"] = refs

        if assignment is not None:
            os.chdir(cwd)

    def _generate_impl_report(self, refs, group=None):
        nb = nbformat.read(self.submission_path, as_version=NBFORMAT_VERSION)
        self._remove_plugin_calls(nb)

        stu = StudentImplementation(nb)
        # # if self.plugin_config.get("student_cache_dir") is not None:
        #     fn = os.path.splitext(os.path.basename(self.submission_path))[0] + ".pkl"
        #     fp = self._cache_student_impl(stu, fn)
        #     cache = f"Student implementation cached at: {fp}\n"
        # else:
        #     cache = ""
        
        results = stu.check(refs, group=group)

        correct = [r.correct for r in results]
        
        report = f"PyBryt Reference Messages:\n"
        
        if any(correct):
            results = [results[correct.index(True)]]
            ending = "\nREFERENCE SATISFIED"

        else:
            ending = "\nNO REFERENCE SATISFIED"

        # TODO: separate mssages for separate references
        for m in itertools.chain(*(r.messages for r in results)):
            report += f"- {m}\n"
        
        report += ending

        self._generated_report = report
        self._student_impl = stu
    
    def _cache_student_impl(self, results: GradingResults, stu: StudentImplementation) -> NoReturn:
        # path = self.plugin_config.get("student_cache_dir")
        # if path is None or not os.path.isabs(path):
        #     raise ValueError(f"Invalid student implementation cache directory: {path}")
        # elif not os.path.isdir(path):
        #     raise ValueError(f"Student implementation cache directory does not exist or is a file: {path}")
        # fn = filename
        # fp = os.path.join(path, fn)
        # i = 1
        # while os.path.exists(fp):
        #     stem, ext = os.path.splitext(filename)
        #     fn = stem + f"-{i}" + ext
        #     fp = os.path.join(path, fn)
        #     i += 1
        
        # stu.dump(dest=fp)
        encoded_stu = stu.dumps()
        data = results.get_plugin_data(self.IMPORTABLE_NAME, {})
        data["cached_student_impl"] = encoded_stu
        results.set_plugin_data(self.IMPORTABLE_NAME, data)

    @classmethod
    def _remove_plugin_calls(cls, nb):
        """
        """
        for cell in nb['cells']:
            source = get_source(cell)
            for i in range(len(source)):
                if f".run_plugin(\"{cls.IMPORTABLE_NAME}\"" in source[i] or f".add_plugin_files(\"{cls.IMPORTABLE_NAME}\"" in source[i]:
                    source[i] = "# " + source[i]
            cell["source"] = "\n".join(source)

    def from_notebook(self, *ref_paths, group=None):
        saved = save_notebook(self.submission_path)
        if not saved:
            warnings.warn(
                "Could not force-save notebook; the results of this call will be based on the last "
                "saved version of this notebook."
            )

        refs = []
        for rp in ref_paths:
            ref = ReferenceImplementation.load(rp)
            if isinstance(ref, ReferenceImplementation):
                refs.append(ref)
            else:
                refs.extend(ref)
        
        self._generate_impl_report(refs, group=group)
        print(self._generated_report)
    
    def notebook_export(self, dest="student.pkl"):
        """
        """
        if self._student_impl is not None:
            self._student_impl.dump(dest)
        else:
            raise RuntimeError("Could not find a cached student implementation to export")
        return [dest]

    def before_execution(self, submission):
        self._remove_plugin_calls(submission)
        return submission

    def after_grading(self, results):
        if self._student_impl is None:
            ref_bytes = base64.b64decode(self.plugin_config["reference_bytes"])
            refs = dill.loads(ref_bytes)
            self._generate_impl_report(refs)
        
        self._cache_student_impl(results, self._student_impl)

    def generate_report(self):
        if self._generated_report is None:
            ref_bytes = base64.b64decode(self.plugin_config["reference_bytes"])
            refs = dill.loads(ref_bytes)
            self._generate_impl_report(refs)
        
        return self._generated_report
    
    @classmethod
    def load_cached_implementations(cls, results: Union[GradingResults, List[GradingResults]]) -> \
            Union[StudentImplementation, List[StudentImplementation]]:
        """
        """
        if isinstance(results, GradingResults):
            data = results.get_plugin_data(cls.IMPORTABLE_NAME)["cached_student_impl"]
            return StudentImplementation.loads(data)
        elif isinstance(results, list):
            if not all(isinstance(r, GradingResults) for r in results):
                raise TypeError("Object of unexpected type received in results")
            data = [r.get_plugin_data(cls.IMPORTABLE_NAME) for r in results]
            return [StudentImplementation.loads(d["cached_student_impl"]) for d in data]
        else:
            raise TypeError("invalid type for argument 'results'")



"""
"""

import dill
import click

from .reference import ReferenceImplementation

@click.group()
def main():
    pass

@main.command("compile")
@click.argument("src", type=click.Path(exists=True))
@click.argument("dest", type=click.Path(), default="reference.pkl")
def compile_reference(src, dest):
    refs = ReferenceImplementation.compile(src)
    with open(dest, "wb+") as f:
        dill.dump(refs, f)



"""
"""

import json
import random
import string
import dill
import hashlib
import time
import nbformat

from typing import Any, List, NoReturn, Union
from IPython import get_ipython
from IPython.display import display, Javascript


def pickle_and_hash(obj: Any) -> str:
    s = dill.dumps(obj)
    return hashlib.sha512(s).hexdigest()


def filter_pickleable_list(lst: List[Any]) -> NoReturn:
    """
    """

    to_delete = []
    for i, v in enumerate(lst):
        try:
            dill.dumps(v)
        except:
            to_delete.append(i)
    
    to_delete.reverse()
    for i in to_delete:
        lst.pop(i)


def notebook_to_string(nb_path: Union[str, nbformat.NotebookNode]) -> str:
    """
    """
    
    if isinstance(nb_path, str):
        with open(nb_path) as f:
            nb = json.load(f)
    elif isinstance(nb_path, nbformat.NotebookNode):
        nb = nb_path
    else:
        raise TypeError("invalid notebook type")
    
    source = ""
    for cell in nb['cells']:
        if cell['cell_type'] == 'code':
            if isinstance(cell['source'], list):
                source += "".join(cell['source']) + "\n"
            else:
                assert isinstance(cell['source'], str), f"could not parse notebook cell: {cell}"
                source += cell['source'] + "\n"

    source = "\n".join(l for l in source.split("\n") if not l.startswith("%") and not l.startswith("!"))
    return source


def make_secret(size=6, chars=string.ascii_uppercase + string.digits):
    """
    Used to generate a dynamic variable name for grading functions

    This function generates a random name using the given length and character set.
    
    Args:
        size (``int``): length of output name
        chars (``str``, optional): set of characters used to create function name
    
    Returns:
        ``str``: randomized string name for grading function
    """
    return ''.join(random.choice(chars) for _ in range(size))


def save_notebook(filename, timeout=10):
    """
    Force-saves a Jupyter notebook by displaying JavaScript.

    Args:
        filename (``str``): path to notebook file being saved
        timeout (``int`` or ``float``): number of seconds to wait for save before timing-out
    
    Returns
        ``bool``: whether the notebook was saved successfully
    """
    if get_ipython() is not None:
        f = open(filename, "rb")
        md5 = hashlib.md5(f.read()).hexdigest()
        start = time.time_ns()
        display(Javascript("Jupyter.notebook.save_checkpoint();"))
        
        curr = md5
        while curr == md5 and time.time_ns() - start <= timeout * 10**9:
            time.sleep(1)
            f.seek(0)
            curr = hashlib.md5(f.read()).hexdigest()
        
        f.close()

        return curr != md5



"""Annotations for marking-up reference implementations"""

from . import invariants
from .annotation import Annotation, AnnotationResult
# from .function import Function
from .relation import *
from .value import *



"""
"""

import os
import dill
import json
import warnings
import numpy as np

from copy import deepcopy
from typing import Any, List, NoReturn, Optional, Tuple, Union

from .annotations import Annotation, AnnotationResult
from .utils import notebook_to_string

class ReferenceImplementation:
    
    annotations: List[Annotation]

    def __init__(self, annotations: List[Annotation]):
        if not isinstance(annotations, list):
            raise TypeError("annotations should be a list of Annotations")
        if not all(isinstance(ann, Annotation) for ann in annotations):
            raise TypeError("Found non-annotation in annotations")
        
        self.annotations = annotations

    @staticmethod
    def load(file: str) -> Union[List['ReferenceImplementation'], 'ReferenceImplementation']:
        with open(file, "rb") as f:
            instance = dill.load(f)
        return instance

    def run(self, observed_values: List[Tuple[Any, float]], group: Optional[str] = None) -> 'ReferenceResult':        
        if group is not None:
            annots = []
            for ann in self.annotations:
                if ann.group == group:
                    annots.append(ann)
            if len(annots) == 0:
                raise ValueError(f"Group '{group}' not found")
        
        else:
            annots = self.annotations
        
        results = []
        for exp in annots:
            results.append(exp.check(observed_values))
        
        return ReferenceResult(self, results, group=group)

    def dump(self, dest: str = "reference.pkl") -> NoReturn:
        with open(dest, "wb+") as f:
            dill.dump(self, f)

    @classmethod
    def compile(cls, file: str) -> Union['ReferenceImplementation', List['ReferenceImplementation']]:
        Annotation.reset_tracked_annotations()

        ext = os.path.splitext(file)[1]
        if ext == ".ipynb":
            source = notebook_to_string(file)

        else:
            with open(file) as f:
                source = f.read()

        env = {}
        exec(source, env)

        refs = []
        annots = []
        for _, v in env.items():
            if isinstance(v, cls):
                refs.append(v)
            # if isinstance(v, Annotation):
            #     annots.append(v)

        if not refs:
            if not Annotation.get_tracked_annotations():
                warnings.warn(f"Could not find any reference implementations in {file}")
            else:
                refs = [cls(deepcopy(Annotation.get_tracked_annotations()))]

        Annotation.reset_tracked_annotations()
        
        if len(refs) == 1:
            return refs[0]
        return refs


class ReferenceResult:
    """
    """

    reference: ReferenceImplementation
    results: List[AnnotationResult]
    group: Optional[str]

    def __init__(self, reference, annotation_results: List[AnnotationResult], group=None):
        self.reference = reference
        self.results = annotation_results
        self.group = group
    
    def __repr__(self):
        results = ',\n  '.join(repr(r) for r in self.results)
        return f"ReferenceResult([\n  {results}\n])"

    @property
    def correct(self):
        return all(r.satisfied for r in self.results)

    @property
    def messages(self):
        messages = []
        message_names = {}
        for r in self.results:
            m = r.messages
            for msg in m:
                if msg[1] is None:
                    messages.append(msg[0])
                else:
                    if msg[1] in message_names and not msg[2]:
                        idx = message_names[msg[1]]
                        messages[idx] = msg[0]
                    elif msg[1] not in message_names:
                        message_names[msg[1]] = len(messages)
                        messages.append(msg[0])
        return messages

    def to_array(self):
        """
        """
        return np.array([r.satisfied for r in self.results], dtype=int)
